// NOTE NOTE NOTE:
//
//   The configuration parameters are at the bottom of this file.
//
//

// *** Constants.

outlets = 2;
inlets = 1;

// DMX channels.
_SIX_INCAND = 0;
_THREE_FLUOR = 1;
_FOUR_RED = 2;


// *** Classes.

function _StateMachine(initial_cue, transitions, delays, cue_triggers) {
	this.initial_cue = initial_cue;
	this.transitions = transitions;
	this.delays = delays;
	this.cue_triggers = cue_triggers;
	this.task = null;
	this.substate = null;

	// Normalize the transition table.
	for (var i in this.transitions) {
		var tr = this.transitions[i];
		var sum = 0;
		for (var j in tr)
			sum += tr[j];
		if (sum) {
			for (var j in tr)
				tr[j] /= sum;
		}
	}

	this.next = function() {
		this.set_cue(this.transition(this.cue));
	};

	// Go directly to a given cue.
	this.set_cue = function(cue) {
		this.stop();
		this.cue = cue;
		this.cue_triggers[cue]();

		var time = this.delays[cue]();
		outlet(1, time);

		this.task = new Task(this.next, this, []);
		this.task.schedule(time * 60000);
	};

	this.stop = function() {
		if (this.task) {
			this.task.cancel();
			this.task = null;
		}
		if (this.substate)
			this.substate.stop();
	};

	this.start = function() {
		this.set_cue(this.initial_cue);
	};

	// Given a cue, return the next cue, using the _TRANSITION_TABLE to select
	// probabilities.
	this.transition = function(cue) {
		var probs = this.transitions[cue];
		var total_prob = Math.random();
		for (var i in probs) {
			total_prob -= probs[i];
			if (total_prob <= 0)
				return i;
		}

		throw RangeError;
	};
};

// *** Local functions.

function _dmx(messages) {
	for (m in messages)
		outlet(0, messages[m]);
};


// Return a random Poisson variable, distributed around a mean.
function _poisson(mean) {
	return mean * -Math.log(Math.random());
};


// *** Global variable.

var _state_machine;

// *** Global functions.

function start() {
	_state_machine.stop();
	_state_machine.start();
};

function next() {
	if (_state_machine)
		_state_machine.next();
};

function stop() {
	if (_state_machine)
		_state_machine.stop();
};

function append(name) {
	if (name != 'None')
		outlet(0, name);
};

function init() {
	outlet(0, 'menu');
};

function clear() { };
function set() {};

// *****************************************************
// ***                                               ***
// ***      Configuration parameters start here!     ***
// ***                                               ***
// *****************************************************

var _state_machine = new _StateMachine
	('default_cue',  // Initial cue.

	 // These next tables determine what happens for each cue.  You can just
	 // make up new cues at will but each cue has to have an entry in all three
	 // tables below or errors will occur.  Order of the cues is unimportant
	 // but might as well keep it alphabetic.

	 // The probability of transitioning from a given cue to another cue.
	 // You should try to make sure that the total probabily in each line is
	 // exactly 100% or else the distribution will be somewhat wrong.
	 {
	 default_cue: { fluor_cue:1,   incand_cue:1, flicker_cue:1, },
	 fluor_cue:   { default_cue:2, incand_cue:1, flicker_cue:1, },
	 incand_cue:  { default_cue:2, fluor_cue:1,  flicker_cue:1, },
	 flicker_cue: { default_cue:2, fluor_cue:1,  incand_cue:1, },
	 },

	 // How long to stay in a given cue, in minutes.
	 //
	 // Each entry in the table is a function that's called afresh each time we
	 // enter the cue.  If you wanted a uniform distribution with mean 5 minutes
	 // and standard deviation 3 minutes, you'd say
	 //
	 //   cue_new: function() { return _uniform(5, 3); },
	 //
	 {
	 default_cue: function() { return _poisson(0.003); },
   fluor_cue: function() { return _poisson(0.005); },
   incand_cue: function() { return _poisson(0.005); },
	 flicker_cue: function() { return _poisson(0.005); },
   },

	 // For each cue, a list of operations messages emitted upon entry.
	 {
	 default_cue: function() {
			 _dmx([[_SIX_INCAND, 255], [_THREE_FLUOR, 0],   [_FOUR_RED, 0]]);
		 },
	 fluor_cue: function() {
			 _dmx([[_SIX_INCAND, 0],   [_THREE_FLUOR, 255], [_FOUR_RED, 0]]);
		 },
	 incand_cue: function() {
			 _dmx([[_SIX_INCAND, 0],   [_THREE_FLUOR, 0],   [_FOUR_RED, 255]]);
		 },
	 flicker_cue: function() {
			 _flicker_state_machine.start();
		 }
	 });

var _flicker_state_machine = new _StateMachine
	('on',
	 {  // Transitions.
		 on: { off:1, },
		 off: { on:1, },
	 },

	 {  // Delay times.
		 on:  function() { return _poisson(0.007); },
		 off: function() { return _poisson(0.003); },
	 },

	 // For each cue, a list of operations messages emitted upon entry.
	 {
		 on: function() {
			 _dmx([[_SIX_INCAND, 255], [_THREE_FLUOR, 0],   [_FOUR_RED, 0]]);
		 },
		 off: function() {
			 _dmx([[_SIX_INCAND, 0], [_THREE_FLUOR, 0],   [_FOUR_RED, 0]]);
		 },
	 });

_state_machine.substate = _flicker_state_machine;
